@@ -9,8 +9,8 @@
    Expressions can be compiled into ast node for later evaluation,
    using the values in the symbol table current at evaluation time.
 
-   version: 0.9.13
-   last update: 2018-Sept-29
+   version: 0.9.27
+   last update: 2022-Feb-05
    License:  MIT
    Author:  Matthew Newville <newville@cars.uchicago.edu>
             Center for Advanced Radiation Sources,
@@ -21,7 +21,7 @@ from .asteval import Interpreter
 from .astutils import (NameFinder, get_ast_names, make_symbol_table,
                        valid_symbol_name)
 
-__all__ = ['Interpreter', 'NameFinder', 'valid_symbol_name',
-           'make_symbol_table', 'get_ast_names']
+from .version import version as __version__
 
-from ._version import __version__
+__all__ = ['Interpreter', 'NameFinder', 'valid_symbol_name',
+           'make_symbol_table', 'get_ast_names', '__version__']
