@@ -37,6 +37,7 @@ functions that are considered unsafe are missing ('eval', 'exec', and
 'getattr' for example)
 """
 import ast
+import copy
 import inspect
 import time
 from sys import exc_info, stderr, stdout
@@ -69,6 +70,8 @@ class Interpreter:
         callable file-like object where standard error will be sent.
     use_numpy : bool
         whether to use functions from numpy.
+    max_statement_length : int
+        maximum length of expression allowed [50,000 characters]
     minimal : bool
         create a minimal interpreter: disable all options (see Note 1).
     no_if : bool
@@ -107,8 +110,9 @@ class Interpreter:
     """
 
     def __init__(self, symtable=None, usersyms=None, writer=None,
-                 err_writer=None, use_numpy=True, minimal=False,
-                 no_if=False, no_for=False, no_while=False, no_try=False,
+                 err_writer=None, use_numpy=True,
+                 max_statement_length=50000, minimal=False, no_if=False,
+                 no_for=False, no_while=False, no_try=False,
                  no_functiondef=False, no_ifexp=False, no_listcomp=False,
                  no_augassign=False, no_assert=False, no_delete=False,
                  no_raise=False, no_print=False, max_time=None,
@@ -116,6 +120,7 @@ class Interpreter:
 
         self.writer = writer or stdout
         self.err_writer = err_writer or stderr
+        self.max_statement_length = max(1, min(1.e8, max_statement_length))
 
         if symtable is None:
             if usersyms is None:
@@ -128,6 +133,7 @@ class Interpreter:
         self.error_msg = None
         self.expr = None
         self.retval = None
+        self._calldepth = 0
         self.lineno = 0
         self.start_time = time.time()
         self.use_numpy = HAS_NUMPY and use_numpy
@@ -179,8 +185,8 @@ class Interpreter:
 
         self.no_deepcopy = [key for key, val in symtable.items()
                             if (callable(val)
-                                or 'numpy.lib.index_tricks' in repr(val)
-                                or inspect.ismodule(val))]
+                                or inspect.ismodule(val)
+                                or 'numpy.lib.index_tricks' in repr(type(val)))]
 
     def remove_nodehandler(self, node):
         """remove support for a node
@@ -233,7 +239,7 @@ class Interpreter:
         self._interrupt = ast.Raise()
         self.error.append(err)
         if self.error_msg is None:
-            self.error_msg = "at expr='%s'" % (self.expr)
+            self.error_msg = (' '.join([msg, "at expr='%s'" % (self.expr)])).strip()
         elif len(msg) > 0:
             self.error_msg = msg
         if exc is None:
@@ -249,6 +255,10 @@ class Interpreter:
     #  eval:   string statement -> result = run(parse(statement))
     def parse(self, text):
         """Parse statement/expression to Ast representation."""
+        if len(text) > self.max_statement_length:
+            msg = f'length of text exceeds {self.max_statement_length:d} characters'
+            self.raise_exception(None, msg='Runtime Error', expr=msg)
+
         self.expr = text
         try:
             out = ast.parse(text)
@@ -309,21 +319,29 @@ class Interpreter:
         self.lineno = lineno
         self.error = []
         self.start_time = time.time()
-        try:
-            node = self.parse(expr)
-        except:
-            errmsg = exc_info()[1]
-            if len(self.error) > 0:
-                errmsg = "\n".join(self.error[0].get_error())
-            if raise_errors:
-                try:
-                    exc = self.error[0].exc
-                except:
-                    exc = RuntimeError
-                raise exc(errmsg)
-            if show_errors:
-                print(errmsg, file=self.err_writer)
-            return
+        if isinstance(expr, str):
+            if len(expr) > self.max_statement_length:
+                msg = f'length of text exceeds {self.max_statement_length:d} characters'
+                raise ValueError(msg)
+
+            try:
+                node = self.parse(expr)
+            except:
+                errmsg = exc_info()[1]
+                if len(self.error) > 0:
+                    errmsg = "\n".join(self.error[0].get_error())
+                if raise_errors:
+                    try:
+                        exc = self.error[0].exc
+                    except:
+                        exc = RuntimeError
+                    raise exc(errmsg)
+                if show_errors:
+                    print(errmsg, file=self.err_writer)
+                return
+        else:
+            node = expr
+
         try:
             return self.run(node, expr=expr, lineno=lineno)
         except:
@@ -356,6 +374,8 @@ class Interpreter:
 
     def on_return(self, node):  # ('value',)
         """Return statement: look for None, return special sentinel."""
+        if self._calldepth == 0:
+            raise SyntaxError('cannot return at top level')
         self.retval = self.run(node.value)
         if self.retval is None:
             self.retval = ReturnedNone
@@ -656,19 +676,80 @@ class Interpreter:
         self._interrupt = None
 
     def on_listcomp(self, node):    # ('elt', 'generators')
-        """List comprehension."""
+        """List comprehension -- only up to 4 generators!"""
         out = []
+        locals = {}
+        saved_syms = {}
+
         for tnode in node.generators:
             if tnode.__class__ == ast.comprehension:
+                if tnode.target.__class__ == ast.Name:
+                    if (not valid_symbol_name(tnode.target.id) or
+                        tnode.target.id in self.readonly_symbols):
+                        errmsg = "invalid symbol name (reserved word?) %s" % tnode.target.id
+                        self.raise_exception(tnode.target, exc=NameError, msg=errmsg)
+                    locals[tnode.target.id] = []
+                    if tnode.target.id in self.symtable:
+                        saved_syms[tnode.target.id] = copy.deepcopy(self.symtable[tnode.target.id])
+
+                elif tnode.target.__class__ == ast.Tuple:
+                    target = []
+                    for tval in tnode.target.elts:
+                        locals[tval.id] = []
+                        if tval.id in self.symtable:
+                            saved_syms[tval.id] = copy.deepcopy(self.symtable[tval.id])
+
+        for tnode in node.generators:
+            if tnode.__class__ == ast.comprehension:
+                tlist = []
+                ttype = 'name'
+                if tnode.target.__class__ == ast.Name:
+                    if (not valid_symbol_name(tnode.target.id) or
+                        tnode.target.id in self.readonly_symbols):
+                        errmsg = "invalid symbol name (reserved word?) %s" % tnode.target.id
+                        self.raise_exception(tnode.target, exc=NameError, msg=errmsg)
+                    ttype, target = 'name', tnode.target.id
+                elif tnode.target.__class__ == ast.Tuple:
+                    ttype = 'tuple'
+                    target =tuple([tval.id for tval in tnode.target.elts])
+
                 for val in self.run(tnode.iter):
-                    self.node_assign(tnode.target, val)
+                    if ttype == 'name':
+                        self.symtable[target] = val
+                    else:
+                        for telem, tval in zip(target, val):
+                            self.symtable[target] = val
+
                     add = True
                     for cond in tnode.ifs:
                         add = add and self.run(cond)
                     if add:
-                        out.append(self.run(node.elt))
+                        if ttype == 'name':
+                            locals[target].append(val)
+                        else:
+                            for telem, tval in zip(target, val):
+                                locals[telem].append(tval)
+
+        def listcomp_recurse(i, names, data):
+            if i == len(names):
+                out.append(self.run(node.elt))
+                return
+
+            for val in data[i]:
+                self.symtable[names[i]] = val
+                listcomp_recurse(i+1, names, data)
+
+        names = list(locals.keys())
+        data = list(locals.values())
+
+        listcomp_recurse(0, names, data)
+
+        for name, val in saved_syms.items():
+            self.symtable[name] = val
+
         return out
 
+
     def on_excepthandler(self, node):  # ('type', 'name', 'body')
         """Exception handler..."""
         return (self.run(node.type), node.name, node.body)
@@ -746,13 +827,20 @@ class Interpreter:
         if kwargs is not None:
             keywords.update(self.run(kwargs))
 
+        if isinstance(func, Procedure):
+            self._calldepth += 1
         try:
-            return func(*args, **keywords)
+            out = func(*args, **keywords)
         except Exception as ex:
+            out = None
             func_name = getattr(func, '__name__', str(func))
             self.raise_exception(
                 node, msg="Error running function call '%s' with args %s and "
                 "kwargs %s: %s" % (func_name, args, keywords, ex))
+        finally:
+            if isinstance(func, Procedure):
+                self._calldepth -= 1
+        return out
 
     def on_arg(self, node):    # ('test', 'msg')
         """Arg for function definitions."""
@@ -929,6 +1017,7 @@ class Procedure:
         save_symtable = self.__asteval__.symtable.copy()
         self.__asteval__.symtable.update(symlocals)
         self.__asteval__.retval = None
+        self.__asteval__._calldepth += 1
         retval = None
 
         # evaluate script of function
@@ -944,5 +1033,6 @@ class Procedure:
                 break
 
         self.__asteval__.symtable = save_symtable
+        self.__asteval__._calldepth -= 1
         symlocals = None
         return retval
